/*
 * bulkdata.c - Bulkdata core
 *
 * Copyright (C) 2022, IOPSYS Software Solutions AB.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 * Author: Omar Kallel <omar.kallel@pivasoftware.com>
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <string.h>
#include <libubox/uloop.h>

#include "config.h"
#include "utils.h"

LIST_HEAD(failedreports);

static void bulkdata_profile_cb(struct uloop_timeout *timeout);

static int get_retry_period(int min)
{
	srand(time(NULL));
	return rand()%min + min;
}

static void bulkdata_run_profiles(struct bulkdata *bulkdata)
{
	for (int i = 0; i < bulkdata->profile_number; i++) {
		bulkdata->profile[i].utimer.cb = bulkdata_profile_cb;
		
		bulkdata->profile[i].failed_reports = &failedreports;
		unsigned int next_period = get_next_period(bulkdata->profile[i].time_reference, bulkdata->profile[i].reporting_interval);
		INFO("The session of profile name '%s' will be start in %d sec", bulkdata->profile[i].profile_name, next_period);
		uloop_timeout_set(&bulkdata->profile[i].utimer, next_period * 1000);
	}
}

static void bulkdata_profile_cb(struct uloop_timeout *timeout)
{
	struct profile *profile = NULL;
	unsigned int http_code = 0;
	char *report = NULL;

	profile = container_of(timeout, struct profile, utimer);
	time_t now = time(NULL);

	INFO("New session of profile name '%s' started", profile->profile_name);
	if (profile->retry_count == 0 || profile->next_retry > now || !profile->http_retry_enable) //Perdiodic execution
		create_encoding_bulkdata_report(profile, &report);
	else
		create_failed_report(profile, &report);

	DEBUG("The content of the profile name report '%s' is :\n==========\n%s\n==========\n", profile->profile_name, report);
	http_code = http_send_message(profile, report);
	if (http_code != 200) {
		unsigned int retry_period = 0;

		if (profile->retry_count == 0 || profile->next_retry > now || !profile->http_retry_enable) {  //Perdiodic execution
			retry_period = get_retry_period(profile->http_retry_minimum_wait_interval);
			profile->next_period = now + profile->reporting_interval;
			profile->next_retry = now + retry_period;
			profile->retry_count = 1;
			profile->min_retry = profile->http_retry_minimum_wait_interval * 2;
			if ((profile->next_retry < profile->next_period) && profile->http_retry_enable) {
				INFO("Retry session of profile name '%s' in %d sec", profile->profile_name, retry_period);
				uloop_timeout_set(timeout, 1000 * retry_period);
			} else {
				INFO("Start New session of profile name '%s' in %d sec", profile->profile_name, profile->reporting_interval);
				uloop_timeout_set(timeout, 1000 * profile->reporting_interval);
			}
		} else { //Retry execution
			retry_period = get_retry_period(profile->min_retry);
			profile->min_retry*=2;
			profile->next_retry+=retry_period;
			profile->retry_count++;
			if (profile->next_retry < profile->next_period) {
				INFO("Retry session of profile name '%s' in %d sec", profile->profile_name, retry_period);
				uloop_timeout_set(timeout, 1000 * retry_period);
			} else {
				INFO("Retry session of profile name '%s' in %ld sec", profile->profile_name, (profile->next_period-profile->next_retry+retry_period));
				uloop_timeout_set(timeout, 1000 * (profile->next_period-profile->next_retry+retry_period));
			}
		}

		if (profile->new_report) {
			bulkdata_add_failed_report(profile, profile->new_report);
			FREE(profile->new_report);
		}
		FREE(report);
	} else {
		if (profile->retry_count == 0 || profile->next_retry > now || !profile->http_retry_enable) {
			INFO("Start New session of profile name '%s' in %d sec", profile->profile_name, profile->reporting_interval);
			uloop_timeout_set(timeout, 1000 * profile->reporting_interval);
		} else {
			INFO("Retry session of profile name '%s' in %ld sec", profile->profile_name, (profile->next_period-profile->next_retry));
			uloop_timeout_set(timeout, 1000 * (profile->next_period-profile->next_retry));
		}
		FREE(profile->new_report);
		FREE(report);
		empty_failed_reports_list(profile);
		profile->retry_count= 0;
	}
}

int main(void)
{
	struct bulkdata bulkdata_main = {0};
	int error = 0;

	/* Logging to syslog */
	openlog("bulkdata", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	INFO("Start bulkdatad daemon");
	error = wait_for_usp_raw_object();
	if (error)
		return error;

	bulkdata_config_init(&bulkdata_main);
	
	uloop_init();
	bulkdata_run_profiles(&bulkdata_main);
	uloop_run();

	uloop_done();
	bulkdata_config_fini(&bulkdata_main);
	INFO("Stop bulkdatad daemon");
	closelog();
	return 0;
}
